#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A simple example showcasing how to use pywhispercpp to transcribe a recording.
"""
import argparse
import logging
import sounddevice as sd
import pywhispercpp.constants
from pywhispercpp.model import Model
import importlib.metadata


__version__ = importlib.metadata.version('pywhispercpp')

__header__ = f"""
===================================================================
PyWhisperCpp
A simple example of transcribing a recording, based on whisper.cpp
Version: {__version__}               
===================================================================
"""


class Recording:
    """
    Recording class

    Example usage
    ```python
    from pywhispercpp.examples.recording import Recording

    myrec = Recording(5)
    myrec.start()
    ```
    """
    def __init__(self,
                 duration: int,
                 model: str = 'tiny.en',
                 **model_params):
        self.duration = duration
        self.sample_rate = pywhispercpp.constants.WHISPER_SAMPLE_RATE
        self.channels = 1
        self.pwcpp_model = Model(model, print_realtime=True, **model_params)

    def start(self):
        logging.info(f"Start recording for {self.duration}s ...")
        recording = sd.rec(int(self.duration * self.sample_rate), samplerate=self.sample_rate, channels=self.channels)
        sd.wait()
        logging.info('Duration finished')
        res = self.pwcpp_model.transcribe(recording)
        self.pwcpp_model.print_timings()


def _main():
    print(__header__)
    parser = argparse.ArgumentParser(description="", allow_abbrev=True)
    # Positional args
    parser.add_argument('duration', type=int, help=f"duration in seconds")
    parser.add_argument('-m', '--model', default='tiny.en', type=str, help="Whisper.cpp model, default to %(default)s")

    args = parser.parse_args()

    myrec = Recording(duration=args.duration, model=args.model)
    myrec.start()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A simple example showcasing the use of `pywhispercpp` as an assistant.
The idea is to use a `VAD` to detect speech (in this example we used webrtcvad), and when speech is detected
we run the inference.
"""
import argparse
import importlib.metadata
import queue
import time
from typing import Callable
import numpy as np
import sounddevice as sd
import pywhispercpp.constants as constants
import webrtcvad
import logging
from pywhispercpp.model import Model

__version__ = importlib.metadata.version('pywhispercpp')

__header__ = f"""
=====================================
PyWhisperCpp
A simple assistant using Whisper.cpp
Version: {__version__}               
=====================================
"""


class Assistant:
    """
    Assistant class

    Example usage
    ```python
    from pywhispercpp.examples.assistant import Assistant

    my_assistant = Assistant(commands_callback=print, n_threads=8)
    my_assistant.start()
    ```
    """

    def __init__(self,
                 model='tiny',
                 input_device: int = None,
                 silence_threshold: int = 8,
                 q_threshold: int = 16,
                 block_duration: int = 30,
                 commands_callback: Callable[[str], None] = None,
                 **model_params):

        """
        :param model: whisper.cpp model name or a direct path to a`ggml` model
        :param input_device: The input device (aka microphone), keep it None to take the default
        :param silence_threshold: The duration of silence after which the inference will be running
        :param q_threshold: The inference won't be running until the data queue is having at least `q_threshold` elements
        :param block_duration: minimum time audio updates in ms
        :param commands_callback: The callback to run when a command is received
        :param model_log_level: Logging level
        :param model_params: any other parameter to pass to the whsiper.cpp model see ::: pywhispercpp.constants.PARAMS_SCHEMA
        """

        self.input_device = input_device
        self.sample_rate = constants.WHISPER_SAMPLE_RATE  # same as whisper.cpp
        self.channels = 1  # same as whisper.cpp
        self.block_duration = block_duration
        self.block_size = int(self.sample_rate * self.block_duration / 1000)
        self.q = queue.Queue()

        self.vad = webrtcvad.Vad()
        self.silence_threshold = silence_threshold
        self.q_threshold = q_threshold
        self._silence_counter = 0

        self.pwccp_model = Model(model,
                                 print_realtime=False,
                                 print_progress=False,
                                 print_timestamps=False,
                                 single_segment=True,
                                 no_context=True,
                                 **model_params)
        self.commands_callback = commands_callback

    def _audio_callback(self, indata, frames, time, status):
        """
        This is called (from a separate thread) for each audio block.
        """
        if status:
            logging.warning(F"underlying audio stack warning:{status}")

        assert frames == self.block_size
        audio_data = map(lambda x: (x + 1) / 2, indata)  # normalize from [-1,+1] to [0,1]
        audio_data = np.fromiter(audio_data, np.float16)
        audio_data = audio_data.tobytes()
        detection = self.vad.is_speech(audio_data, self.sample_rate)
        if detection:
            self.q.put(indata.copy())
            self._silence_counter = 0
        else:
            if self._silence_counter >= self.silence_threshold:
                if self.q.qsize() > self.q_threshold:
                    self._transcribe_speech()
                    self._silence_counter = 0
            else:
                self._silence_counter += 1

    def _transcribe_speech(self):
        logging.info(f"Speech detected ...")
        audio_data = np.array([])
        while self.q.qsize() > 0:
            # get all the data from the q
            audio_data = np.append(audio_data, self.q.get())
        # Appending zeros to the audio data as a workaround for small audio packets (small commands)
        audio_data = np.concatenate([audio_data, np.zeros((int(self.sample_rate) + 10))])
        # running the inference
        self.pwccp_model.transcribe(audio_data,
                                    new_segment_callback=self._new_segment_callback)

    def _new_segment_callback(self, seg):
        if self.commands_callback:
            self.commands_callback(seg.text)

    def start(self) -> None:
        """
        Use this function to start the assistant
        :return: None
        """
        logging.info(f"Starting Assistant ...")
        with sd.InputStream(
                device=self.input_device,  # the default input device
                channels=self.channels,
                samplerate=constants.WHISPER_SAMPLE_RATE,
                blocksize=self.block_size,
                callback=self._audio_callback):

            try:
                logging.info(f"Assistant is listening ... (CTRL+C to stop)")
                while True:
                    time.sleep(0.1)
            except KeyboardInterrupt:
                logging.info("Assistant stopped")

    @staticmethod
    def available_devices():
        return sd.query_devices()


def _main():
    parser = argparse.ArgumentParser(description="", allow_abbrev=True)
    # Positional args
    parser.add_argument('-m', '--model', default='tiny.en', type=str, help="Whisper.cpp model, default to %(default)s")
    parser.add_argument('-ind', '--input_device', type=int, default=None,
                        help=f'Id of The input device (aka microphone)\n'
                             f'available devices {Assistant.available_devices()}')
    parser.add_argument('-st', '--silence_threshold', default=16, type=int,
                        help=f"he duration of silence after which the inference will be running, default to %(default)s")
    parser.add_argument('-bd', '--block_duration', default=30,
                        help=f"minimum time audio updates in ms, default to %(default)s")

    args = parser.parse_args()

    my_assistant = Assistant(model=args.model,
                             input_device=args.input_device,
                             silence_threshold=args.silence_threshold,
                             block_duration=args.block_duration,
                             commands_callback=print)
    my_assistant.start()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A simple Command Line Interface to test the package
"""
import argparse
import importlib.metadata
import logging

import pywhispercpp.constants as constants

__version__ = importlib.metadata.version('pywhispercpp')

__header__ = f"""
PyWhisperCpp
A simple Command Line Interface to test the package
Version: {__version__}               
====================================================
"""

from pywhispercpp.model import Model
import pywhispercpp.utils as utils


def _get_params(args) -> dict:
    """
    Helper function to get params from argparse as a `dict`
    """
    params = {}
    for arg in args.__dict__:
        if arg in constants.PARAMS_SCHEMA.keys() and getattr(args, arg) is not None:
            if constants.PARAMS_SCHEMA[arg]['type'] is bool:
                if getattr(args, arg).lower() == 'false':
                    params[arg] = False
                else:
                    params[arg] = True
            else:
                params[arg] = constants.PARAMS_SCHEMA[arg]['type'](getattr(args, arg))
    return params


def run(args):
    logging.info(f"Running with model `{args.model}`")
    params = _get_params(args)
    logging.info(f"Running with params {params}")
    m = Model(model=args.model, **params)
    logging.info(f"System info: n_threads = {m.get_params()['n_threads']} | Processors = {args.processors} "
                 f"| {m.system_info()}")
    for file in args.media_file:
        segs = []
        try:
            logging.info(f"Processing file {file} ...")
            m.transcribe(file,
                         n_processors=int(args.processors) if args.processors else None,
                         new_segment_callback=lambda seg: segs.append(seg)
                         )
            m.print_timings()
        except KeyboardInterrupt:
            logging.info("Transcription manually stopped")
            break
        except Exception as e:
            logging.error(f"Error while processing file {file}: {e}")
        finally:
            if segs:
                # output stuff
                if args.output_txt:
                    logging.info(f"Saving result as a txt file ...")
                    txt_file = utils.output_txt(segs, file)
                    logging.info(f"txt file saved to {txt_file}")
                if args.output_vtt:
                    logging.info(f"Saving results as a vtt file ...")
                    vtt_file = utils.output_vtt(segs, file)
                    logging.info(f"vtt file saved to {vtt_file}")
                if args.output_srt:
                    logging.info(f"Saving results as a srt file ...")
                    srt_file = utils.output_srt(segs, file)
                    logging.info(f"srt file saved to {srt_file}")
                if args.output_csv:
                    logging.info(f"Saving results as a csv file ...")
                    csv_file = utils.output_csv(segs, file)
                    logging.info(f"csv file saved to {csv_file}")


def main():
    print(__header__)
    parser = argparse.ArgumentParser(description="", allow_abbrev=True)
    # Positional args
    parser.add_argument('media_file', type=str, nargs='+', help="The path of the media file or a list of files"
                                                                "separated by space")

    parser.add_argument('-m', '--model', default='tiny', help="Path to the `ggml` model, or just the model name")

    parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')
    parser.add_argument('--processors', help="number of processors to use during computation")
    parser.add_argument('-otxt', '--output-txt', action='store_true', help="output result in a text file")
    parser.add_argument('-ovtt', '--output-vtt', action='store_true', help="output result in a vtt file")
    parser.add_argument('-osrt', '--output-srt', action='store_true', help="output result in a srt file")
    parser.add_argument('-ocsv', '--output-csv', action='store_true', help="output result in a CSV file")

    # add params from PARAMS_SCHEMA
    for param in constants.PARAMS_SCHEMA:
        param_fields = constants.PARAMS_SCHEMA[param]
        parser.add_argument(f'--{param}',
                            help=f'{param_fields["description"]}')

    args = parser.parse_args()
    run(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Quick and dirty realtime livestream transcription.

Not fully satisfying though :)
You are welcome to make it better.
"""
import argparse
import logging
import queue
from multiprocessing import Process
import ffmpeg
import numpy as np
import pywhispercpp.constants as constants
import sounddevice as sd
from pywhispercpp.model import Model
import importlib.metadata


__version__ = importlib.metadata.version('pywhispercpp')

__header__ = f"""
========================================================
PyWhisperCpp
A simple Livestream transcription, based on whisper.cpp
Version: {__version__}               
========================================================
"""

class LiveStream:
    """
    LiveStream class

    ???+ note

        It heavily depends on the machine power, the processor will jump quickly to 100% with the wrong parameters.

    Example usage
    ```python
    from pywhispercpp.examples.livestream import LiveStream

    url = ""  # Make sure it is a direct stream URL
    ls = LiveStream(url=url, n_threads=4)
    ls.start()
    ```
    """

    def __init__(self,
                 url,
                 model='tiny.en',
                 block_size: int = 1024,
                 buffer_size: int = 20,
                 sample_size: int = 4,
                 output_device: int = None,
                 model_log_level=logging.CRITICAL,
                 **model_params):

        """
        :param url: Live stream url <a direct stream URL>
        :param model: whisper.cpp model
        :param block_size: block size, default to 1024
        :param buffer_size: number of blocks used for buffering, default to 20
        :param sample_size: sample size
        :param output_device: the output device, aka the speaker, leave it None to take the default
        :param model_log_level: logging level
        :param model_params: any other whisper.cpp params
        """
        self.url = url
        self.block_size = block_size
        self.buffer_size = buffer_size
        self.sample_size = sample_size
        self.output_device = output_device

        self.channels = 1
        self.samplerate = constants.WHISPER_SAMPLE_RATE

        self.q = queue.Queue(maxsize=buffer_size)
        self.audio_data = np.array([])

        self.pwccp_model = Model(model,
                                 log_level=model_log_level,
                                 print_realtime=True,
                                 print_progress=False,
                                 print_timestamps=False,
                                 single_segment=True,
                                 **model_params)

    def _transcribe_process(self):
        self.pwccp_model.transcribe(self.audio_data, n_processors=None)

    def _audio_callback(self, outdata, frames, time, status):
        assert frames == self.block_size
        if status.output_underflow:
            logging.error('Output underflow: increase blocksize?')
            raise sd.CallbackAbort
        assert not status
        try:
            data = self.q.get_nowait()
        except queue.Empty as e:
            logging.error('Buffer is empty: increase buffer_size?')
            raise sd.CallbackAbort from e
        assert len(data) == len(outdata)
        outdata[:] = data
        audio = np.frombuffer(data[:], np.float32)
        audio = audio.reshape((audio.size, 1)) / 2 ** 5
        self.audio_data = np.append(self.audio_data, audio)
        if self.audio_data.size > self.samplerate:
            # Create a separate process for transcription
            p1 = Process(target=self._transcribe_process,)
            p1.start()
            self.audio_data = np.array([])

    def start(self):
        process = ffmpeg.input(self.url).output(
            'pipe:',
            format='f32le',
            acodec='pcm_f32le',
            ac=self.channels,
            ar=self.samplerate,
            loglevel='quiet',
        ).run_async(pipe_stdout=True)

        out_stream = sd.RawOutputStream(
            device=self.output_device,
            samplerate=self.samplerate,
            blocksize=self.block_size,
            channels=self.channels,
            dtype='float32',
            callback=self._audio_callback)

        read_size = self.block_size * self.channels * self.sample_size

        logging.info('Buffering ...')
        for _ in range(self.buffer_size):
            self.q.put_nowait(process.stdout.read(read_size))

        with out_stream:
            logging.info('Starting Playback ... (CTRL+C) to stop')
            try:
                timeout = self.block_size * self.buffer_size / self.samplerate
                while True:
                    buffer_data = process.stdout.read(read_size)
                    self.q.put(buffer_data, timeout=timeout)
            except KeyboardInterrupt:
                logging.info("Interrupted!")

    @staticmethod
    def available_devices():
        return sd.query_devices()


def _main():
    print(__header__)
    parser = argparse.ArgumentParser(description="", allow_abbrev=True)
    # Positional args
    parser.add_argument('url', type=str, help=f"Stream URL")

    parser.add_argument('-nt', '--n_threads', type=int, default=3,
                        help="number of threads, default to %(default)s")
    parser.add_argument('-m', '--model', default='tiny.en', type=str, help="Whisper.cpp model, default to %(default)s")
    parser.add_argument('-od', '--output_device', type=int, default=None,
                        help=f'the output device, aka the speaker, leave it None to take the default\n'
                             f'available devices {LiveStream.available_devices()}')
    parser.add_argument('-bls', '--block_size', type=int, default=1024,
                        help=f"block size, default to %(default)s")
    parser.add_argument('-bus', '--buffer_size', type=int, default=20,
                        help=f"number of blocks used for buffering, default to %(default)s")
    parser.add_argument('-ss', '--sample_size', type=int, default=4,
                        help=f"Sample size, default to %(default)s")
    args = parser.parse_args()


    # url = "http://n03.radiojar.com/t2n88q0st5quv?rj-ttl=5&rj-tok=AAABhsR2u6MAYFxz69dJ6eQnww"  # VOA english
    ls = LiveStream(url=args.url,
                    model=args.model,
                    block_size=args.block_size,
                    buffer_size=args.buffer_size,
                    sample_size=args.sample_size,
                    output_device=args.output_device,
                    n_threads=args.n_threads)
    ls.start()


if __name__ == '__main__':
    _main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains a simple Python API on-top of the C-style
[whisper.cpp](https://github.com/ggerganov/whisper.cpp) API.
"""
import importlib.metadata
import logging
import shutil
import sys
from pathlib import Path
from time import time
from typing import Union, Callable, List, TextIO, Tuple, Optional
import _pywhispercpp as pw
import numpy as np
import pywhispercpp.utils as utils
import pywhispercpp.constants as constants
import subprocess
import os
import tempfile
import wave

__author__ = "absadiki"
__copyright__ = "Copyright 2023, "
__license__ = "MIT"
__version__ = importlib.metadata.version('pywhispercpp')

logger = logging.getLogger(__name__)


class Segment:
    """
    A small class representing a transcription segment
    """

    def __init__(self, t0: int, t1: int, text: str, probability: float = np.nan):
        """
        :param t0: start time
        :param t1: end time
        :param text: text
        :param probability: Confidence score for the segment, computed as the geometric mean of
            the token probabilities for the segment (NaN if not calculated).
            This makes it interpretable as a probability in [0, 1].
        """
        self.t0 = t0
        self.t1 = t1
        self.text = text
        self.probability = probability

    def __str__(self):
        return f"t0={self.t0}, t1={self.t1}, text={self.text}, probability={self.probability}"

    def __repr__(self):
        return str(self)


class Model:
    """
    This classes defines a Whisper.cpp model.

    Example usage.
    ```python
    model = Model('base.en', n_threads=6)
    segments = model.transcribe('file.mp3')
    for segment in segments:
        print(segment.text)
    ```
    """

    _new_segment_callback = None

    def __init__(self,
                 model: str = 'tiny',
                 models_dir: str = None,
                 params_sampling_strategy: int = 0,
                 redirect_whispercpp_logs_to: Union[bool, TextIO, str, None] = False,
                 **params):
        """
        :param model: The name of the model, one of the [AVAILABLE_MODELS](/pywhispercpp/#pywhispercpp.constants.AVAILABLE_MODELS),
                        (default to `tiny`), or a direct path to a `ggml` model.
        :param models_dir: The directory where the models are stored, or where they will be downloaded if they don't
                            exist, default to [MODELS_DIR](/pywhispercpp/#pywhispercpp.constants.MODELS_DIR) <user_data_dir/pywhsipercpp/models>
        :param params_sampling_strategy: 0 -> GREEDY, else BEAM_SEARCH
        :param redirect_whispercpp_logs_to: where to redirect the whisper.cpp logs, default to False (no redirection), accepts str file path, sys.stdout, sys.stderr, or use None to redirect to devnull
        :param params: keyword arguments for different whisper.cpp parameters,
                        see [PARAMS_SCHEMA](/pywhispercpp/#pywhispercpp.constants.PARAMS_SCHEMA)
        """
        if Path(model).is_file():
            self.model_path = model
        else:
            self.model_path = utils.download_model(model, models_dir)
        self._ctx = None
        self._sampling_strategy = pw.whisper_sampling_strategy.WHISPER_SAMPLING_GREEDY if params_sampling_strategy == 0 else \
            pw.whisper_sampling_strategy.WHISPER_SAMPLING_BEAM_SEARCH
        self._params = pw.whisper_full_default_params(self._sampling_strategy)
        # assign params
        self._set_params(params)
        self.redirect_whispercpp_logs_to = redirect_whispercpp_logs_to
        # init the model
        self._init_model()

    def transcribe(self,
                   media: Union[str, np.ndarray],
                   n_processors: int = None,
                   new_segment_callback: Callable[[Segment], None] = None,
                   **params) -> List[Segment]:
        """
        Transcribes the media provided as input and returns list of `Segment` objects.
        Accepts a media_file path (audio/video) or a raw numpy array.

        :param media: Media file path or a numpy array
        :param n_processors: if not None, it will run the transcription on multiple processes
                             binding to whisper.cpp/whisper_full_parallel
                             > Split the input audio in chunks and process each chunk separately using whisper_full()
        :param new_segment_callback: callback function that will be called when a new segment is generated
        :param params: keyword arguments for different whisper.cpp parameters, see ::: constants.PARAMS_SCHEMA
        :param extract_probability: If True, calculates the geometric mean of token probabilities for each segment,
            providing a confidence score interpretable as a probability in [0, 1].
        :return: List of transcription segments
        """
        if type(media) is np.ndarray:
            audio = media
        else:
            if not Path(media).exists():
                raise FileNotFoundError(media)
            audio = self._load_audio(media)

        # Handle extract_probability parameter
        self.extract_probability = params.pop('extract_probability', False)

        # update params if any
        self._set_params(params)

        # setting up callback
        if new_segment_callback:
            Model._new_segment_callback = new_segment_callback
            pw.assign_new_segment_callback(self._params, Model.__call_new_segment_callback)

        # run inference
        start_time = time()
        logger.info("Transcribing ...")
        res = self._transcribe(audio, n_processors=n_processors)
        end_time = time()
        logger.info(f"Inference time: {end_time - start_time:.3f} s")
        return res

    @staticmethod
    def _get_segments(ctx, start: int, end: int, extract_probability: bool = False) -> List[Segment]:
        """
        Helper function to get generated segments between `start` and `end`

        :param ctx: whisper context
        :param start: start index
        :param end: end index
        :param extract_probability: whether to calculate token probabilities

        :return: list of segments
        """
        n = pw.whisper_full_n_segments(ctx)
        assert end <= n, f"{end} > {n}: `End` index must be less or equal than the total number of segments"
        res = []
        for i in range(start, end):
            t0 = pw.whisper_full_get_segment_t0(ctx, i)
            t1 = pw.whisper_full_get_segment_t1(ctx, i)
            bytes = pw.whisper_full_get_segment_text(ctx, i)
            text = bytes.decode('utf-8', errors='replace')

            avg_prob = np.nan

            # Only calculate probabilities if requested
            if extract_probability:
                n_tokens = pw.whisper_full_n_tokens(ctx, i)
                if n_tokens == 1:
                    avg_prob = pw.whisper_full_get_token_p(ctx, i, 0)
                elif n_tokens > 1:
                    total_logprob = 0.0
                    for j in range(n_tokens):
                        total_logprob += np.log(pw.whisper_full_get_token_p(ctx, i, j))
                    avg_prob = np.exp(total_logprob / n_tokens)
                else:
                    avg_prob = np.nan

            res.append(Segment(t0, t1, text.strip(), probability=np.float32(avg_prob)))
        return res

    def get_params(self) -> dict:
        """
        Returns a `dict` representation of the actual params

        :return: params dict
        """
        res = {}
        for param in dir(self._params):
            if param.startswith('__'):
                continue
            try:
                res[param] = getattr(self._params, param)
            except Exception:
                # ignore callback functions
                continue
        return res

    @staticmethod
    def get_params_schema() -> dict:
        """
        A simple link to ::: constants.PARAMS_SCHEMA
        :return: dict of params schema
        """
        return constants.PARAMS_SCHEMA

    @staticmethod
    def lang_max_id() -> int:
        """
        Returns number of supported languages.
        Direct binding to whisper.cpp/lang_max_id
        :return:
        """
        return pw.whisper_lang_max_id()

    def print_timings(self) -> None:
        """
        Direct binding to whisper.cpp/whisper_print_timings

        :return: None
        """
        pw.whisper_print_timings(self._ctx)

    @staticmethod
    def system_info() -> None:
        """
        Direct binding to whisper.cpp/whisper_print_system_info

        :return: None
        """
        return pw.whisper_print_system_info()

    @staticmethod
    def available_languages() -> list[str]:
        """
        Returns a list of supported language codes

        :return: list of supported language codes
        """
        n = pw.whisper_lang_max_id()
        res = []
        for i in range(n+1):
            res.append(pw.whisper_lang_str(i))
        return res

    def _init_model(self) -> None:
        """
        Private method to initialize the method from the bindings, it will be called automatically from the __init__
        :return:
        """
        logger.info("Initializing the model ...")
        with utils.redirect_stderr(to=self.redirect_whispercpp_logs_to):
            self._ctx = pw.whisper_init_from_file(self.model_path)

    def _set_params(self, kwargs: dict) -> None:
        """
        Private method to set the kwargs params to the `Params` class
        :param kwargs: dict like object for the different params
        :return: None
        """
        for param in kwargs:
            setattr(self._params, param, kwargs[param])

    def _transcribe(self, audio: np.ndarray, n_processors: int = None):
        """
        Private method to call the whisper.cpp/whisper_full function

        :param audio: numpy array of audio data
        :param n_processors: if not None, it will run whisper.cpp/whisper_full_parallel with n_processors
        :return:
        """

        if n_processors:
            pw.whisper_full_parallel(self._ctx, self._params, audio, audio.size, n_processors)
        else:
            pw.whisper_full(self._ctx, self._params, audio, audio.size)
        n = pw.whisper_full_n_segments(self._ctx)
        res = Model._get_segments(self._ctx, 0, n, self.extract_probability)
        return res

    @staticmethod
    def __call_new_segment_callback(ctx, n_new, user_data) -> None:
        """
        Internal new_segment_callback, it just calls the user's callback with the `Segment` object
        :param ctx: whisper.cpp ctx param
        :param n_new: whisper.cpp n_new param
        :param user_data: whisper.cpp user_data param
        :return: None
        """
        n = pw.whisper_full_n_segments(ctx)
        start = n - n_new
        res = Model._get_segments(ctx, start, n, False)
        for segment in res:
            Model._new_segment_callback(segment)

    @staticmethod
    def _load_audio(media_file_path: str) -> np.array:
        """
         Helper method to return a `np.array` object from a media file
         If the media file is not a WAV file, it will try to convert it using ffmpeg

        :param media_file_path: Path of the media file
        :return: Numpy array
        """

        def wav_to_np(file_path):
            with wave.open(file_path, 'rb') as wf:
                num_channels = wf.getnchannels()
                sample_width = wf.getsampwidth()
                sample_rate = wf.getframerate()
                num_frames = wf.getnframes()

                if num_channels not in (1, 2):
                    raise Exception(f"WAV file must be mono or stereo")

                if sample_rate != pw.WHISPER_SAMPLE_RATE:
                    raise Exception(f"WAV file must be {pw.WHISPER_SAMPLE_RATE} Hz")

                if sample_width != 2:
                    raise Exception(f"WAV file must be 16-bit")

                raw = wf.readframes(num_frames)
                wf.close()
                audio = np.frombuffer(raw, dtype=np.int16).astype(np.float32)
                n = num_frames
                if num_channels == 1:
                    pcmf32 = audio / 32768.0
                else:
                    audio = audio.reshape(-1, 2)
                    # Averaging the two channels
                    pcmf32 = (audio[:, 0] + audio[:, 1]) / 65536.0
                return pcmf32

        if media_file_path.endswith('.wav'):
            return wav_to_np(media_file_path)
        else:
            if shutil.which('ffmpeg') is None:
                raise Exception(
                    "FFMPEG is not installed or not in PATH. Please install it, or provide a WAV file or a NumPy array instead!")

            temp_file = tempfile.NamedTemporaryFile(suffix=".wav", delete=False)
            temp_file_path = temp_file.name
            temp_file.close()
            try:
                subprocess.run([
                    'ffmpeg', '-i', media_file_path, '-ac', '1', '-ar', '16000',
                    temp_file_path, '-y'
                ], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                return wav_to_np(temp_file_path)
            finally:
                os.remove(temp_file_path)

    def auto_detect_language(self,  media: Union[str, np.ndarray], offset_ms: int = 0, n_threads: int = 4) -> Tuple[Tuple[str, np.float32], dict[str, np.float32]]:
        """
        Automatic language detection using whisper.cpp/whisper_pcm_to_mel and whisper.cpp/whisper_lang_auto_detect

        :param media: Media file path or a numpy array
        :param offset_ms: offset in milliseconds
        :param n_threads: number of threads to use
        :return: ((detected_language, probability), probabilities for all languages)
        """
        if type(media) is np.ndarray:
            audio = media
        else:
            if not Path(media).exists():
                raise FileNotFoundError(media)
            audio = self._load_audio(media)

        pw.whisper_pcm_to_mel(self._ctx, audio, len(audio), n_threads)
        lang_max_id = self.lang_max_id()
        probs = np.zeros(lang_max_id, dtype=np.float32)
        auto_detect = pw.whisper_lang_auto_detect(self._ctx, offset_ms, n_threads, probs)
        langs = self.available_languages()
        lang_probs = {langs[i]: probs[i] for i in range(lang_max_id)}
        return (langs[auto_detect], probs[auto_detect]), lang_probs

    def __del__(self):
        """
        Free up resources
        :return: None
        """
        pw.whisper_free(self._ctx)


import sys
import threading
from datetime import datetime

from PyQt5.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QHBoxLayout, QPushButton,
    QFileDialog, QProgressBar, QLabel, QFrame,
    QSizePolicy, QTableWidget, QTableWidgetItem, QHeaderView,
    QGroupBox, QFormLayout, QComboBox, QLineEdit, QCheckBox,
    QSpinBox, QDoubleSpinBox, QToolButton, QDialog, QMenu  # Import QMenu
)
from PyQt5.QtCore import Qt, pyqtSignal, QObject
import os
import importlib.metadata

__version__ = importlib.metadata.version('pywhispercpp')

from pywhispercpp.model import Model, Segment
from pywhispercpp.utils import output_txt, output_srt, output_vtt, output_csv  # Import utilities


# --- Available Models ---
# Define a custom order for model sizes
MODEL_SIZE_ORDER = {"tiny": 0, "base": 1, "small": 2, "medium": 3, "large": 4}

UNSORTED_MODELS = [
    "base", "base-q5_1", "base-q8_0", "base.en", "base.en-q5_1",
    "base.en-q8_0", "large-v1", "large-v2", "large-v2-q5_0",
    "large-v2-q8_0", "large-v3", "large-v3-q5_0", "large-v3-turbo",
    "large-v3-turbo-q5_0", "large-v3-turbo-q8_0", "medium",
    "medium-q5_0", "medium-q8_0", "medium.en", "medium.en-q5_0",
    "medium.en-q8_0", "small", "small-q5_1", "small-q8_0", "small.en",
    "small.en-q5_1", "small.en-q8_0", "tiny", "tiny-q5_1", "tiny-q8_0",
    "tiny.en", "tiny.en-q5_1", "tiny.en-q8_0",
]


# Custom sort key function
def get_model_sort_key(model_name):
    # Extract the base model name (e.g., "tiny.en" -> "tiny", "large-v3-turbo" -> "large")
    base_name = model_name.split('.')[0].split('-')[0]
    # Return a tuple for multi-level sorting: (size_order, full_model_name_for_secondary_sort)
    return (MODEL_SIZE_ORDER.get(base_name, 99), model_name)  # 99 for any unexpected names


# Sort the models
AVAILABLE_MODELS = sorted(UNSORTED_MODELS, key=get_model_sort_key)

# --- Retouched Minimal Stylesheet ---
STYLESHEET = """
/* General Application Styles */
QWidget {
    background-color: #f5f5f5; /* Light gray background */
    color: #333333; /* Dark text */
    font-family: Arial, sans-serif;
    font-size: 14px;
}

/* Buttons */
QPushButton {
    background-color: #e0e0e0; /* Slightly darker gray for buttons */
    color: #333333;
    border: 1px solid #c0c0c0; /* Light gray border */
    padding: 6px 12px;
    border-radius: 3px; /* Slightly rounded corners */
    outline: none; /* Remove focus outline for a cleaner look */
}

QPushButton:hover {
    background-color: #d0d0d0; /* Darker on hover */
}

QPushButton:pressed {
    background-color: #c0c0c0; /* Even darker when pressed */
}

QPushButton:disabled {
    background-color: #f0f0f0;
    color: #aaaaaa;
    border-color: #e0e0e0;
}

/* Specific styling for the Transcribe button (Call to Action) */
QPushButton#TranscribeButton {
    background-color: #007bff; /* Vibrant blue */
    color: #ffffff; /* White text */
    font-weight: bold;
    font-size: 15px; /* Slightly larger font */
    padding: 8px 18px; /* More padding */
    border: 1px solid #007bff; /* Matching border */
    border-radius: 4px;
}

QPushButton#TranscribeButton:hover {
    background-color: #0056b3; /* Darker blue on hover */
    border-color: #0056b3;
}

QPushButton#TranscribeButton:pressed {
    background-color: #004085; /* Even darker blue when pressed */
    border-color: #004085;
}

QPushButton#TranscribeButton:disabled {
    background-color: #cccccc; /* Light gray for disabled state */
    color: #888888;
    border-color: #cccccc;
}

/* Stop button */
QPushButton#StopButton {
    background-color: #ff0000; /* Red for stop */
    color: #ffffff; /* White text */
    font-weight: bold;
    font-size: 15px; /* Slightly larger font */
    padding: 8px 18px; /* More padding */
    border: 1px solid #ff0000; /* Matching border */
    border-radius: 4px;
}

QPushButton#StopButton:hover {
    background-color: #cc0000; /* Darker red on hover */
    border-color: #cc0000;
}

QPushButton#StopButton:pressed {
    background-color: #990000; /* Even darker red when pressed */
    border-color: #990000;
}

QPushButton#StopButton:disabled {
    background-color: #cccccc; /* Light gray for disabled state */
    color: #888888;
    border-color: #cccccc;
}

/* ToolButton for accordion header */
QToolButton {
    background-color: transparent; /* Keep it transparent */
    border: none;
    padding: 5px 0; /* Some padding for clickable area */
    font-weight: bold;
    color: #444444; /* Darker text for emphasis */
    text-align: left;
    outline: none; /* Remove focus outline */
}

QToolButton::menu-indicator {
    image: none; /* Hide default menu indicator */
}

/* Table Widget */
QTableWidget {
    background-color: #ffffff;
    border: 1px solid #d0d0d0; /* Slightly softer border */
    gridline-color: #e8e8e8; /* Very light grid lines */
    border-radius: 3px;
}

QTableWidget::item {
    padding: 4px;
    border-bottom: 1px solid #f5f5f5; /* Match general background for subtle row separation */
}

/* Table Header */
QHeaderView::section {
    background-color: #e8e8e8; /* Light header background */
    color: #333333;
    padding: 5px;
    border: 1px solid #d0d0d0;
    font-weight: bold;
}

/* Labels */
QLabel {
    color: #333333;
}

/* Specific styling for the main title label */
QLabel#TitleLabel {
    font-size: 18px; /* Slightly smaller than previous '20px' to integrate better */
    font-weight: bold;
    color: #0056b3; /* A slightly darker blue for main title */
    padding-bottom: 3px;
    margin-bottom: 5px;
    border-bottom: 1px solid #e0e0e0;
}

/* Progress Bar */
QProgressBar {
    border: 1px solid #bbbbbb;
    border-radius: 3px;
    text-align: center;
    background-color: #e6e6e6;
    color: #333333;
}

QProgressBar::chunk {
    background-color: #4CAF50; /* A pleasant green */
    border-radius: 2px;
}

/* Input Widgets (ComboBox, LineEdit, SpinBox, DoubleSpinBox) */
QComboBox, QLineEdit, QSpinBox, QDoubleSpinBox {
    border: 1px solid #cccccc;
    padding: 3px;
    background-color: #ffffff;
    color: #333333;
    border-radius: 3px; /* Apply rounded corners */
}

/* File label should visually match text inputs */
QLabel#file_label {
    border: 1px solid #cccccc;
    padding: 3px;
    background-color: #ffffff;
    border-radius: 3px;
}

QFrame {
    border: none; /* Keep frames invisible unless needed for structure */
}

/* Status Bar Label */
QLabel#status_bar_label {
    background-color: #e0e0e0; /* Light grey background */
    border-top: 1px solid #cccccc; /* Separator */
    color: #444444; /* Darker text */
    padding: 3px 5px; /* Add internal padding */
    font-size: 13px; /* Slightly smaller font */
}
"""


# --- Communication Object for Threading ---
class WorkerSignals(QObject):
    """
    Defines signals available from a running worker thread.
    Supported signals are:
    - finished: No data
    - error: tuple (exctype, value, traceback.format_exc())
    - result: list (the transcribed segments)
    - progress: int (0-100)
    - status_update: str
    """
    finished = pyqtSignal()
    error = pyqtSignal(tuple)
    segment = pyqtSignal(Segment)
    result = pyqtSignal(list)
    progress = pyqtSignal(int)
    status_update = pyqtSignal(str)


# --- Worker Thread for Transcription ---
class PyWhisperCppWorker(threading.Thread):

    def __init__(self, audio_file_path, model_name, **transcribe_params):
        super().__init__()
        self.audio_file_path = audio_file_path
        self.model_name = model_name
        self.transcribe_params = transcribe_params
        self.signals = WorkerSignals()
        self._is_running = False

    def run(self):
        """
        Executes the transcription process.
        """
        try:
            self._is_running = True
            self.signals.status_update.emit(f"Loading model: {self.model_name}...")

            # pywhispercpp will download the specified model if not found
            model_init_params = {}
            if 'n_threads' in self.transcribe_params and self.transcribe_params['n_threads'] is not None:
                model_init_params['n_threads'] = self.transcribe_params['n_threads']
                # Remove from transcribe_params as it's a model init param
                del self.transcribe_params['n_threads']

            model = Model(self.model_name, **model_init_params)

            self.signals.status_update.emit("Model loaded. Starting transcription...")

            def new_segment_callback(segment):
                if not self._is_running:
                    raise RuntimeError("Transcription manually stopped")
                self.signals.segment.emit(segment)

            segments = model.transcribe(self.audio_file_path,
                                        new_segment_callback=new_segment_callback,
                                        progress_callback=lambda progress: self.signals.progress.emit(progress),
                                        **self.transcribe_params)

            self.signals.status_update.emit("Transcription complete!")
            self.signals.result.emit(segments)

        except Exception as e:
            print(e)
            self.signals.status_update.emit(f"Error: {str(e)}")
            self.signals.error.emit((type(e), e, str(e)))
        finally:
            self._is_running = False
            self.signals.finished.emit()

    def stop(self):
        self._is_running = False


# --- Main Application Window ---
class TranscriptionApp(QWidget):
    def __init__(self):
        super().__init__()
        self.selected_file_path = None
        self.whisper_thread = None
        # Settings widgets
        self.model_combo = None
        self.language_input = None
        self.translate_checkbox = None
        self.n_threads_spinbox = None
        self.no_context_checkbox = None
        self.temperature_spinbox = None
        self.settings_content_frame = None  # Frame to hold collapsible settings
        self.toggle_settings_button = None  # Button to toggle settings
        self.status_bar_label = None  # New label for the status bar
        self.about_button = None  # About button
        self.segments = []  # Store segments for export
        self.copy_text_button = None  # New button for copy text

        self.initUI()

    def initUI(self):
        """
        Initializes the user interface of the application.
        """
        self.setWindowTitle('PyWhisperCpp Simple GUI')
        self.setGeometry(100, 100, 450, 500)
        # Apply the updated stylesheet
        self.setStyleSheet(STYLESHEET)

        # Main vertical layout
        main_layout = QVBoxLayout()
        # Set bottom margin to 0 for the main layout to ensure status bar is flush
        main_layout.setContentsMargins(4, 4, 4, 0)
        main_layout.setSpacing(10)

        # --- Header (Title + About Button) ---
        header_layout = QHBoxLayout()
        title_label = QLabel("PyWhisperCpp Simple GUI")  # Updated main title label
        title_label.setObjectName("TitleLabel")  # Add objectName for styling
        title_label.setAlignment(Qt.AlignLeft)  # Keep title centered within its allocated space

        # Adding stretch before and after title to center it
        # header_layout.addStretch()
        header_layout.addWidget(title_label)
        header_layout.addStretch()

        # About button
        self.about_button = QPushButton("About")
        self.about_button.clicked.connect(self.show_about_dialog)
        # Removed setFixedSize to allow text to fit, or adjust as needed
        # self.about_button.setFixedSize(50, 25)
        header_layout.addWidget(self.about_button)  # Add it to the header layout

        main_layout.addLayout(header_layout)  # Add the combined header to main layout

        # --- File Selection Area ---
        file_frame = QFrame()
        file_layout = QHBoxLayout(file_frame)
        file_layout.setContentsMargins(0, 0, 0, 0)
        file_layout.setSpacing(10)

        self.select_button = QPushButton("Select Audio File")
        self.select_button.clicked.connect(self.select_file)

        self.file_label = QLabel("No file selected.")
        self.file_label.setObjectName("file_label")  # Added objectName for styling
        self.file_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

        file_layout.addWidget(self.select_button)
        file_layout.addWidget(self.file_label)
        main_layout.addWidget(file_frame)

        # --- Collapsible Settings Section ---
        settings_group = QGroupBox()  # No title here, using QToolButton for title
        settings_group_layout = QVBoxLayout(settings_group)
        settings_group_layout.setContentsMargins(5, 5, 5, 5)

        # Custom title bar for the collapsible group box
        header_layout_settings = QHBoxLayout()  # Renamed to avoid clash
        self.toggle_settings_button = QToolButton(settings_group)
        self.toggle_settings_button.setText("Transcription Settings")
        self.toggle_settings_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toggle_settings_button.setArrowType(Qt.RightArrow)
        self.toggle_settings_button.setCheckable(True)
        self.toggle_settings_button.setChecked(False)  # Start collapsed
        self.toggle_settings_button.clicked.connect(self.toggle_settings_visibility)

        header_layout_settings.addWidget(self.toggle_settings_button)
        header_layout_settings.addStretch()  # Push button to left

        settings_group_layout.addLayout(header_layout_settings)

        # Frame to hold the actual settings form (this will be hidden/shown)
        self.settings_content_frame = QFrame()
        settings_form_layout = QFormLayout(self.settings_content_frame)
        settings_form_layout.setContentsMargins(15, 5, 10, 10)
        settings_form_layout.setSpacing(8)

        # Model Selection
        self.model_combo = QComboBox()
        self.model_combo.addItems(AVAILABLE_MODELS)
        self.model_combo.setCurrentText("tiny")  # Default to 'tiny' as requested
        settings_form_layout.addRow("Model:", self.model_combo)

        # Language Input
        self.language_input = QLineEdit()
        self.language_input.setPlaceholderText('e.g., "en", "es", or leave empty for auto-detect')
        self.language_input.setText("")  # Default to auto-detect
        settings_form_layout.addRow("Language:", self.language_input)

        # Translate Checkbox
        self.translate_checkbox = QCheckBox("Translate to English")
        self.translate_checkbox.setChecked(False)  # Default
        settings_form_layout.addRow("Translate:", self.translate_checkbox)

        # N Threads Spinbox
        self.n_threads_spinbox = QSpinBox()
        self.n_threads_spinbox.setRange(1, os.cpu_count() if os.cpu_count() else 8)  # Max threads based on CPU cores
        self.n_threads_spinbox.setValue(4)  # Sensible default
        settings_form_layout.addRow("Number of Threads:", self.n_threads_spinbox)

        # No Context Checkbox
        self.no_context_checkbox = QCheckBox("No Context (do not use past transcription)")
        self.no_context_checkbox.setChecked(False)  # Default
        settings_form_layout.addRow("No Context:", self.no_context_checkbox)

        # Temperature Spinbox
        self.temperature_spinbox = QDoubleSpinBox()
        self.temperature_spinbox.setRange(0.0, 1.0)
        self.temperature_spinbox.setSingleStep(0.1)
        self.temperature_spinbox.setValue(0.0)  # Default
        settings_form_layout.addRow("Temperature:", self.temperature_spinbox)

        settings_group_layout.addWidget(self.settings_content_frame)
        self.settings_content_frame.setVisible(False)  # Initially hidden

        main_layout.addWidget(settings_group)

        # --- Transcription Button ---
        self.transcribe_button = QPushButton("Transcribe")
        self.transcribe_button.setObjectName("TranscribeButton")  # Add objectName for styling
        self.transcribe_button.setEnabled(False)
        self.transcribe_button.clicked.connect(self.start_transcription)
        main_layout.addWidget(self.transcribe_button)

        # --- Stop Button ---
        self.stop_button = QPushButton("Stop")
        self.stop_button.setObjectName("StopButton")  # Add objectName for styling
        self.stop_button.setEnabled(True)
        self.stop_button.setVisible(False)
        self.stop_button.clicked.connect(self.stop_transcription)
        main_layout.addWidget(self.stop_button)

        # --- Progress Bar ---
        progress_frame = QFrame()
        progress_layout = QVBoxLayout(progress_frame)
        progress_layout.setContentsMargins(0, 5, 0, 5)
        progress_layout.setSpacing(5)

        self.progress_bar = QProgressBar()
        self.progress_bar.setVisible(False)

        progress_layout.addWidget(self.progress_bar)
        main_layout.addWidget(progress_frame)

        # --- Transcription Output Table ---
        output_label = QLabel("Transcription Output:")
        main_layout.addWidget(output_label)

        self.results_table = QTableWidget()
        self.results_table.setColumnCount(3)
        self.results_table.setHorizontalHeaderLabels(["Start Time", "End Time", "Text"])
        header = self.results_table.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.Stretch)
        self.results_table.verticalHeader().setVisible(False)
        main_layout.addWidget(self.results_table)

        # --- Output Buttons (Export and Copy) ---
        output_buttons_layout = QHBoxLayout()
        output_buttons_layout.addStretch()  # Pushes buttons to the right

        # Export Button with Menu
        self.export_button = QPushButton("Export as...")
        self.export_button.setEnabled(False)
        self.export_menu = QMenu(self)

        self.export_action_txt = self.export_menu.addAction("Plain Text (.txt)")
        self.export_action_srt = self.export_menu.addAction("SRT Subtitle (.srt)")
        self.export_action_vtt = self.export_menu.addAction("VTT Subtitle (.vtt)")
        self.export_action_csv = self.export_menu.addAction("CSV (.csv)")

        self.export_action_txt.triggered.connect(lambda: self.export_transcription("txt"))
        self.export_action_srt.triggered.connect(lambda: self.export_transcription("srt"))
        self.export_action_vtt.triggered.connect(lambda: self.export_transcription("vtt"))
        self.export_action_csv.triggered.connect(lambda: self.export_transcription("csv"))

        self.export_button.setMenu(self.export_menu)
        output_buttons_layout.addWidget(self.export_button)

        # Copy Text Button
        self.copy_text_button = QPushButton("Copy Text")
        self.copy_text_button.setEnabled(False)  # Initially disabled
        self.copy_text_button.clicked.connect(self.copy_all_text_to_clipboard)  # Connect to new method
        output_buttons_layout.addWidget(self.copy_text_button)

        main_layout.addLayout(output_buttons_layout)

        # --- Status Bar at the very bottom ---
        self.status_bar_label = QLabel("Ready.")
        self.status_bar_label.setObjectName("status_bar_label")  # Add objectName for styling
        self.status_bar_label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        self.status_bar_label.setContentsMargins(5, 2, 5, 2)
        main_layout.addWidget(self.status_bar_label)

        self.setLayout(main_layout)

    def toggle_settings_visibility(self):
        """Toggles the visibility of the settings content frame and updates the arrow."""
        is_visible = self.settings_content_frame.isVisible()
        self.settings_content_frame.setVisible(not is_visible)
        if not is_visible:
            self.toggle_settings_button.setArrowType(Qt.DownArrow)
        else:
            self.toggle_settings_button.setArrowType(Qt.RightArrow)

    def select_file(self):
        """
        Opens a file dialog to select an audio file.
        """
        options = QFileDialog.Options()
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Select a Media File", "",
            "All Files (*)",
            options=options
        )
        if file_path:
            self.selected_file_path = file_path
            self.file_label.setText(f"Selected: {os.path.basename(file_path)}")
            self.transcribe_button.setEnabled(True)
            self.results_table.setRowCount(0)
            self.export_button.setEnabled(False)  # Disable export until transcription
            self.copy_text_button.setEnabled(False)  # Disable copy until transcription
            self.update_status("File selected: " + os.path.basename(file_path))  # Update new status bar

    def start_transcription(self):
        """
        Starts the transcription process in a separate thread, passing selected settings.
        """
        if self.selected_file_path:
            self.transcribe_button.setVisible(False)
            self.stop_button.setVisible(True)
            self.select_button.setEnabled(False)
            self.progress_bar.setVisible(True)
            self.progress_bar.setValue(0)
            self.results_table.setRowCount(0)
            self.export_button.setEnabled(False)  # Disable export during transcription
            self.copy_text_button.setEnabled(False)  # Disable copy during transcription
            self.update_status("Starting transcription...")
            self.segments = []  # Clear segments for new transcription

            # Gather settings from GUI widgets
            selected_model = self.model_combo.currentText()
            transcribe_params = {
                "language": self.language_input.text() if self.language_input.text() else None,
                "translate": self.translate_checkbox.isChecked(),
                "n_threads": self.n_threads_spinbox.value(),
                "no_context": self.no_context_checkbox.isChecked(),
                "temperature": self.temperature_spinbox.value(),
            }
            # Remove None values to use pywhispercpp defaults where applicable
            transcribe_params = {k: v for k, v in transcribe_params.items() if v is not None}

            # Create and start the worker thread
            self.whisper_thread = PyWhisperCppWorker(
                self.selected_file_path,
                selected_model,
                **transcribe_params
            )
            self.whisper_thread.signals.result.connect(self.on_transcription_result)
            self.whisper_thread.signals.segment.connect(self.on_new_segment)
            self.whisper_thread.signals.finished.connect(self.on_transcription_finished)
            self.whisper_thread.signals.error.connect(self.on_transcription_error)
            self.whisper_thread.signals.progress.connect(self.update_progress)
            self.whisper_thread.signals.status_update.connect(self.update_status)
            self.whisper_thread.start()

    def stop_transcription(self):
        if self.whisper_thread:
            self.whisper_thread.stop()
            # self.transcribe_button.setVisible(True)
            # self.stop_button.setVisible(False)
            # self.select_button.setEnabled(True)
            # self.progress_bar.setVisible(False)
            # self.on_transcription_finished()

    def update_progress(self, value):
        self.progress_bar.setValue(value)
        # Update status bar with progress if not already showing a specific message
        if not self.status_bar_label.text().startswith("Error:") and \
                not self.status_bar_label.text().startswith("Finished.") and \
                not self.status_bar_label.text().startswith("Text exported") and \
                not self.status_bar_label.text().startswith("Text copied"):  # Updated check
            self.update_status(f"Progress: {value}%")

    def update_status(self, status_text):
        # Update the new status bar label directly
        if self.status_bar_label:
            self.status_bar_label.setText(status_text)
            # Polish stylesheet for status_bar_label to ensure updates are reflected
            self.status_bar_label.style().unpolish(self.status_bar_label)
            self.status_bar_label.style().polish(self.status_bar_label)

    def format_time(self, milliseconds):
        """Converts milliseconds to HH:MM:SS.ms format."""
        seconds_total = milliseconds / 1000
        minutes, seconds = divmod(seconds_total, 60)
        hours, minutes = divmod(minutes, 60)
        return f"{int(hours):02d}:{int(minutes):02d}:{seconds:06.3f}"

    def on_new_segment(self, segment):
        row_position = self.results_table.rowCount()
        self.results_table.insertRow(row_position)
        start_time_str = self.format_time(segment.t0)
        end_time_str = self.format_time(segment.t1)

        start_item = QTableWidgetItem(start_time_str)
        end_item = QTableWidgetItem(end_time_str)
        text_item = QTableWidgetItem(segment.text.strip())

        self.results_table.setItem(row_position, 0, start_item)
        self.results_table.setItem(row_position, 1, end_item)
        self.results_table.setItem(row_position, 2, text_item)

    def on_transcription_result(self, segments):
        """
        Populates the results table with the transcription segments.
        Stores segments for export.
        """
        self.segments = segments  # Store segments
        self.export_button.setEnabled(True if segments else False)  # Enable export if segments exist
        self.copy_text_button.setEnabled(True if segments else False)  # Enable copy if segments exist

    def on_transcription_finished(self):
        """
        Cleans up after the transcription thread is finished.
        """
        self.transcribe_button.setVisible(True)
        self.transcribe_button.setEnabled(True)
        self.stop_button.setVisible(False)
        self.select_button.setEnabled(True)
        self.progress_bar.setVisible(False)
        if self.results_table.rowCount() == 0:
            self.update_status("Finished. No transcription data.")
        else:
            self.update_status("Transcription finished successfully!")
        self.whisper_thread = None

    def on_transcription_error(self, err):
        """
        Displays an error message if transcription fails.
        """
        exctype, value, tb = err
        error_message = f"Error: {value}"
        self.update_status(error_message)  # Update new status bar
        self.on_transcription_finished()

    def export_transcription(self, format_type):
        """
        Handles exporting the transcription to a chosen file format.
        """
        if not self.segments:
            self.update_status("No transcription data to export.")
            return

        file_dialog_filter = {
            "txt": "Plain Text Files (*.txt)",
            "srt": "SRT Subtitle Files (*.srt)",
            "vtt": "VTT Subtitle Files (*.vtt)",
            "csv": "CSV (Comma Separated Values) Files (*.csv)",
        }

        default_file_name = os.path.basename(self.selected_file_path).rsplit('.', 1)[
                                0] + f".{format_type}" if self.selected_file_path else f"transcription.{format_type}"

        options = QFileDialog.Options()
        file_path, _ = QFileDialog.getSaveFileName(
            self, f"Save Transcription as {format_type.upper()}",
            default_file_name,
            file_dialog_filter.get(format_type, "All Files (*)"),
            options=options
        )

        if file_path:
            try:
                # Use pywhispercpp.utils functions based on format_type
                if format_type == "txt":
                    # For TXT, we'll re-use the text from the table or segments
                    all_text = []
                    for segment in self.segments:
                        all_text.append(segment.text.strip())
                    output_txt_content = "\n".join(all_text)
                    with open(file_path, 'w', encoding='utf-8') as f:
                        f.write(output_txt_content)

                elif format_type == "srt":
                    if output_srt:
                        output_srt(self.segments, file_path)
                    else:
                        raise ImportError("pywhispercpp.utils.output_srt not available.")
                elif format_type == "vtt":
                    if output_vtt:
                        output_vtt(self.segments, file_path)
                    else:
                        raise ImportError("pywhispercpp.utils.output_vtt not available.")
                elif format_type == "csv":
                    if output_csv:
                        # For CSV, we need to pass a list of lists/tuples representing rows
                        # pywhispercpp.utils.output_csv expects a list of segments and a file path
                        output_csv(self.segments, file_path)
                    else:
                        raise ImportError("pywhispercpp.utils.output_csv not available.")

                self.update_status(f"Transcription successfully exported to {os.path.basename(file_path)}")
            except Exception as e:
                self.update_status(f"Error exporting to {format_type.upper()}: {e}")
        else:
            self.update_status("Export cancelled.")

    def copy_all_text_to_clipboard(self):
        """
        Concatenates all text from segments and copies it to the clipboard.
        """
        if not self.segments:
            self.update_status("No transcription data to copy.")
            return

        all_text = []
        for segment in self.segments:
            all_text.append(segment.text.strip())

        QApplication.clipboard().setText("\n".join(all_text))
        self.update_status("Text copied to clipboard!")

    def show_about_dialog(self):
        """Opens a small dialog with About information."""
        about_dialog = QDialog(self)
        about_dialog.setWindowTitle("About PyWhisperCPP Simple GUI")
        about_dialog.setFixedSize(400, 220)

        dialog_layout = QVBoxLayout(about_dialog)
        dialog_layout.setContentsMargins(20, 20, 20, 20)

        info_text = QLabel()
        info_text.setTextFormat(Qt.RichText)
        info_text.setText(
            "<b>PyWhisperCPP Simple GUI</b><br>"
            f"Version {__version__}<br>"
            "<br>"
            "A simple graphical user interface for PyWhisperCpp Using PyQt.<br><br>"
            "<a href='https://github.com/absadiki/pywhispercpp'>PyWhisperCpp GitHub repository</a><br>"
            "<br>"
            f"Copyright © {datetime.now().year}"
        )
        info_text.setOpenExternalLinks(True)

        dialog_layout.addWidget(info_text)

        close_button = QPushButton("Close")
        close_button.clicked.connect(about_dialog.accept)
        dialog_layout.addWidget(close_button, alignment=Qt.AlignCenter)

        about_dialog.exec_()


def _main():
    """Main function to run the application."""
    if Model is None:
        print("pywhispercpp is not installed.")
        print("Please install it by running: pip install pywhispercpp")
        print("You also need ffmpeg installed on your system.")
        return

    app = QApplication(sys.argv)
    ex = TranscriptionApp()
    ex.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    _main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Constants
"""
from pathlib import Path
from typing import Tuple

import _pywhispercpp as _pwcpp
from platformdirs import user_data_dir


WHISPER_SAMPLE_RATE = _pwcpp.WHISPER_SAMPLE_RATE
# MODELS URL MODELS_BASE_URL+ '/' + MODELS_PREFIX_URL+'-'+MODEL_NAME+'.bin'
# example = "https://huggingface.co/ggerganov/whisper.cpp/resolve/main/ggml-tiny.bin"
MODELS_BASE_URL = "https://huggingface.co/ggerganov/whisper.cpp"
MODELS_PREFIX_URL = "resolve/main/ggml"


PACKAGE_NAME = 'pywhispercpp'


MODELS_DIR = Path(user_data_dir(PACKAGE_NAME)) / 'models'


AVAILABLE_MODELS = [
                "base",
                "base-q5_1",
                "base-q8_0",
                "base.en",
                "base.en-q5_1",
                "base.en-q8_0",
                "large-v1",
                "large-v2",
                "large-v2-q5_0",
                "large-v2-q8_0",
                "large-v3",
                "large-v3-q5_0",
                "large-v3-turbo",
                "large-v3-turbo-q5_0",
                "large-v3-turbo-q8_0",
                "medium",
                "medium-q5_0",
                "medium-q8_0",
                "medium.en",
                "medium.en-q5_0",
                "medium.en-q8_0",
                "small",
                "small-q5_1",
                "small-q8_0",
                "small.en",
                "small.en-q5_1",
                "small.en-q8_0",
                "tiny",
                "tiny-q5_1",
                "tiny-q8_0",
                "tiny.en",
                "tiny.en-q5_1",
                "tiny.en-q8_0",
                ]
PARAMS_SCHEMA = {  # as exactly presented in whisper.cpp
    'n_threads': {
            'type': int,
            'description': "Number of threads to allocate for the inference"
                           "default to min(4, available hardware_concurrency)",
            'options': None,
            'default': None
    },
    'n_max_text_ctx': {
            'type': int,
            'description': "max tokens to use from past text as prompt for the decoder",
            'options': None,
            'default': 16384
    },
    'offset_ms': {
            'type': int,
            'description': "start offset in ms",
            'options': None,
            'default': 0
    },
    'duration_ms': {
            'type': int,
            'description': "audio duration to process in ms",
            'options': None,
            'default': 0
    },
    'translate': {
            'type': bool,
            'description': "whether to translate the audio to English",
            'options': None,
            'default': False
    },
    'no_context': {
            'type': bool,
            'description': "do not use past transcription (if any) as initial prompt for the decoder",
            'options': None,
            'default': False
    },
    'single_segment': {
            'type': bool,
            'description': "force single segment output (useful for streaming)",
            'options': None,
            'default': False
    },
    'print_special': {
            'type': bool,
            'description': "print special tokens (e.g. <SOT>, <EOT>, <BEG>, etc.)",
            'options': None,
            'default': False
    },
    'print_progress': {
            'type': bool,
            'description': "print progress information",
            'options': None,
            'default': True
    },
    'print_realtime': {
            'type': bool,
            'description': "print results from within whisper.cpp (avoid it, use callback instead)",
            'options': None,
            'default': False
    },
    'print_timestamps': {
            'type': bool,
            'description': "print timestamps for each text segment when printing realtime",
            'options': None,
            'default': True
    },
    # [EXPERIMENTAL] token-level timestamps
    'token_timestamps': {
            'type': bool,
            'description': "enable token-level timestamps",
            'options': None,
            'default': False
    },
    'thold_pt': {
            'type': float,
            'description': "timestamp token probability threshold (~0.01)",
            'options': None,
            'default': 0.01
    },
    'thold_ptsum': {
            'type': float,
            'description': "timestamp token sum probability threshold (~0.01)",
            'options': None,
            'default': 0.01
    },
    'max_len': {
            'type': int,
            'description': "max segment length in characters, note: token_timestamps needs to be set to True for this to work",
            'options': None,
            'default': 0
    },
    'split_on_word': {
            'type': bool,
            'description': "split on word rather than on token (when used with max_len)",
            'options': None,
            'default': False
    },
    'max_tokens': {
            'type': int,
            'description': "max tokens per segment (0 = no limit)",
            'options': None,
            'default': 0
    },
    'audio_ctx': {
            'type': int,
            'description': "overwrite the audio context size (0 = use default)",
            'options': None,
            'default': 0
    },
    'initial_prompt': {
                'type': str,
                'description': "Initial prompt, these are prepended to any existing text context from a previous call",
                'options': None,
                'default': None
        },
    'prompt_tokens': {
            'type': Tuple,
            'description': "tokens to provide to the whisper decoder as initial prompt",
            'options': None,
            'default': None
    },
    'prompt_n_tokens': {
            'type': int,
            'description': "tokens to provide to the whisper decoder as initial prompt",
            'options': None,
            'default': 0
    },
    'language': {
            'type': str,
            'description': 'for auto-detection, set to None, "" or "auto"',
            'options': None,
            'default': ""
    },
    'suppress_blank': {
            'type': bool,
            'description': 'common decoding parameters',
            'options': None,
            'default': True
    },
    'suppress_non_speech_tokens': {
            'type': bool,
            'description': 'common decoding parameters',
            'options': None,
            'default': False
    },
    'temperature': {
            'type': float,
            'description': 'initial decoding temperature',
            'options': None,
            'default': 0.0
    },
    'max_initial_ts': {
            'type': float,
            'description': 'max_initial_ts',
            'options': None,
            'default': 1.0
    },
    'length_penalty': {
            'type': float,
            'description': 'length_penalty',
            'options': None,
            'default': -1.0
    },
    'temperature_inc': {
            'type': float,
            'description': 'temperature_inc',
            'options': None,
            'default': 0.2
    },
    'entropy_thold': {
            'type': float,
            'description': 'similar to OpenAI\'s "compression_ratio_threshold"',
            'options': None,
            'default': 2.4
    },
    'logprob_thold': {
            'type': float,
            'description': 'logprob_thold',
            'options': None,
            'default': -1.0
    },
    'no_speech_thold': {  # not implemented
            'type': float,
            'description': 'no_speech_thold',
            'options': None,
            'default': 0.6
    },
    'greedy': {
            'type': dict,
            'description': 'greedy',
            'options': None,
            'default': {"best_of": -1}
    },
    'beam_search': {
            'type': dict,
            'description': 'beam_search',
            'options': None,
            'default': {"beam_size": -1, "patience": -1.0}
    },
    'extract_probability': {
            'type': bool,
            'description': 'calculate the geometric mean of token probabilities for each segment.',
            'options': None,
            'default': True
    }
}



#!/usr/bin/env python

"""
Helper functions
"""

import contextlib
import logging
import os
import sys
from pathlib import Path
from typing import TextIO

import requests
from tqdm import tqdm

from pywhispercpp.constants import (
    AVAILABLE_MODELS,
    MODELS_BASE_URL,
    MODELS_DIR,
    MODELS_PREFIX_URL,
)

logger = logging.getLogger(__name__)


def _get_model_url(model_name: str) -> str:
    """
    Returns the url of the `ggml` model
    :param model_name: name of the model
    :return: URL of the model
    """
    return f"{MODELS_BASE_URL}/{MODELS_PREFIX_URL}-{model_name}.bin"


def download_model(model_name: str, download_dir=None, chunk_size=1024) -> str:
    """
    Helper function to download the `ggml` models
    :param model_name: name of the model, one of ::: constants.AVAILABLE_MODELS
    :param download_dir: Where to store the models
    :param chunk_size: size of the download chunk

    :return: Absolute path of the downloaded model
    """
    if model_name not in AVAILABLE_MODELS:
        logger.error(f"Invalid model name `{model_name}`, available models are: {AVAILABLE_MODELS}")
        return
    if download_dir is None:
        download_dir = MODELS_DIR
        logger.info(f"No download directory was provided, models will be downloaded to {download_dir}")

    os.makedirs(download_dir, exist_ok=True)

    url = _get_model_url(model_name=model_name)
    file_path = Path(download_dir) / os.path.basename(url)
    # check if the file is already there
    if file_path.exists():
        logger.info(f"Model {model_name} already exists in {download_dir}")
    else:
        # download it from huggingface
        resp = requests.get(url, stream=True)
        total = int(resp.headers.get('content-length', 0))

        progress_bar = tqdm(desc=f"Downloading Model {model_name} ...",
                            total=total,
                            unit='iB',
                            unit_scale=True,
                            unit_divisor=1024)

        try:
            with open(file_path, 'wb') as file, progress_bar:
                for data in resp.iter_content(chunk_size=chunk_size):
                    size = file.write(data)
                    progress_bar.update(size)
            logger.info(f"Model downloaded to {file_path.absolute()}")
        except Exception as e:
            # error download, just remove the file
            os.remove(file_path)
            raise e
    return str(file_path.absolute())


def to_timestamp(t: int, separator=',') -> str:
    """
    376 -> 00:00:03,760
    1344 -> 00:00:13,440

    Implementation from `whisper.cpp/examples/main`

    :param t: input time from whisper timestamps
    :param separator: seprator between seconds and milliseconds
    :return: time representation in hh: mm: ss[separator]ms
    """
    # logic exactly from whisper.cpp

    msec = t * 10
    hr = msec // (1000 * 60 * 60)
    msec = msec - hr * (1000 * 60 * 60)
    min = msec // (1000 * 60)
    msec = msec - min * (1000 * 60)
    sec = msec // 1000
    msec = msec - sec * 1000
    return f"{int(hr):02,.0f}:{int(min):02,.0f}:{int(sec):02,.0f}{separator}{int(msec):03,.0f}"


def output_txt(segments: list, output_file_path: str) -> str:
    """
    Creates a raw text from a list of segments

    Implementation from `whisper.cpp/examples/main`

    :param segments: list of segments
    :return: path of the file
    """
    if not output_file_path.endswith('.txt'):
        output_file_path = output_file_path + '.txt'

    absolute_path = Path(output_file_path).absolute()

    with open(str(absolute_path), 'w') as file:
        for seg in segments:
            file.write(seg.text)
            file.write('\n')
    return absolute_path


def output_vtt(segments: list, output_file_path: str) -> str:
    """
    Creates a vtt file from a list of segments

    Implementation from `whisper.cpp/examples/main`

    :param segments: list of segments
    :return: path of the file

    :return: Absolute path of the file
    """
    if not output_file_path.endswith('.vtt'):
        output_file_path = output_file_path + '.vtt'

    absolute_path = Path(output_file_path).absolute()

    with open(absolute_path, 'w') as file:
        file.write("WEBVTT\n\n")
        for seg in segments:
            file.write(f"{to_timestamp(seg.t0, separator='.')} --> {to_timestamp(seg.t1, separator='.')}\n")
            file.write(f"{seg.text}\n\n")
    return absolute_path


def output_srt(segments: list, output_file_path: str) -> str:
    """
    Creates a srt file from a list of segments

    :param segments: list of segments
    :return: path of the file

    :return: Absolute path of the file
    """
    if not output_file_path.endswith('.srt'):
        output_file_path = output_file_path + '.srt'

    absolute_path = Path(output_file_path).absolute()

    with open(absolute_path, 'w') as file:
        for i in range(len(segments)):
            seg = segments[i]
            file.write(f"{i+1}\n")
            file.write(f"{to_timestamp(seg.t0, separator=',')} --> {to_timestamp(seg.t1, separator=',')}\n")
            file.write(f"{seg.text}\n\n")
    return absolute_path


def output_csv(segments: list, output_file_path: str) -> str:
    """
    Creates a srt file from a list of segments

    :param segments: list of segments
    :return: path of the file

    :return: Absolute path of the file
    """
    if not output_file_path.endswith('.csv'):
        output_file_path = output_file_path + '.csv'

    absolute_path = Path(output_file_path).absolute()

    with open(absolute_path, 'w') as file:
        for seg in segments:
            file.write(f"{10 * seg.t0}, {10 * seg.t1}, \"{seg.text}\"\n")
    return absolute_path


@contextlib.contextmanager
def redirect_stderr(to: bool | TextIO | str | None = False) -> None:
    """
    Redirect stderr to the specified target.

    :param to:
        - None to suppress output (redirect to devnull),
        - sys.stdout to redirect to stdout,
        - A file path (str) to redirect to a file,
        - False to do nothing (no redirection).
    """

    if to is False:
        # do nothing
        yield
        return

    def _resolve_target(target):
        opened_stream = None
        if target is None:
            opened_stream = open(os.devnull, "w")
            return opened_stream, True
        if isinstance(target, str):
            opened_stream = open(target, "w")
            return opened_stream, True
        if hasattr(target, "write"):
            return target, False
        raise ValueError(
            "Invalid `to` parameter; expected None, a filepath string, or a file-like object."
        )

    sys.stderr.flush()
    try:
        original_fd = sys.stderr.fileno()
    except (AttributeError, OSError):
        # Jupyter or non-standard stderr implementations
        original_fd = None

    stream, should_close = _resolve_target(to)

    if original_fd is not None and hasattr(stream, "fileno"):
        saved_fd = os.dup(original_fd)
        try:
            os.dup2(stream.fileno(), original_fd)
            yield
        finally:
            os.dup2(saved_fd, original_fd)
            os.close(saved_fd)
            if should_close:
                stream.close()
        return

    # Fallback: Python-level redirect
    try:
        with contextlib.redirect_stderr(stream):
            yield
    finally:
        if should_close:
            stream.close()
